%% ComputeObjectiveFunction
% Myriad objective function for specific sample and parameters.
% 
%% Syntax. 
%%
% 
%   [ Q_of_theta, theta ] = ComputeObjectiveFunction( X );
%   [ Q_of_theta, theta ] = ComputeObjectiveFunction( X , ...);
%   
%% Definitions. 
% For $\gamma \in R$, $w=(w_1, \dots, w_n) \in R^n$ and $X=(x_1, \dots,
% x_n) \in R^n$ the objective function is defined by 
% 
% $$f(\theta) = \sum_{i=1}^n \log(\gamma ^2 + w_i (x_i - \theta)^2)$$
% 
%% Description.
% |[ Q_of_theta, theta ] = ComputeObjectiveFunction( X )| computes the
% objective function for a sample vector |X|. 
% 
% * |theta| represents the values
% where the function is computed. It can be defined in input with the
% option |'Theta'|. 
% * |Q_of_theta| is the objective function computed with values given in
% |theta|.
% 
%% Example.

% Define gamma parameter
gamma=0.03;

% Create a sample vector. 
X= [0.13, 0.86, 0.39, 0.99, 0.27, 0.95, 0.97, 0.16, 0.90];

% Create a weights  vector.
w= [0.70, 0.36, 0.94, 0.22, 0.39, 0.04, 0.26, 0.60, 0.02];

% Compute the objective function. 
[ Q_of_theta, theta ] = ComputeObjectiveFunction( X ,...
                                                     'Weights', w,...
                                                     'Gamma', gamma);

% Display the objective function.
figure, plot(theta, Q_of_theta), title('Objective function.')
%% Input arguments.
% * |'Theta'| represents the values
% where the function is computed. The default parameter is a vector of size
% 1000 which represents the interval from the minimum sample to the maximum
% one, with additionnal margins of size 10% 
%
% Example :
% |[ Q_of_theta, theta ] = ComputeObjectiveFunction( X , 'Theta',
% linspace(0,1,50));|
% 
% *  |'Weights'| represents the weights of the objective function. The size
% must be the same as the one of X. Default values: $\forall i$, 
% $w_i=1$.
% *  |'Gamma'| scalar parameter for $\gamma$. Default value equal to 1. 
%
%
