/*
 *	Myriad.cpp
 *
 *	Updated on:	2017-06-26
 *	   Authors:	Marcel Brusius
 *
 */

#include "Myriad.h"
#include <stdio.h>
#include <mex.h>
#include <math.h>


// tells the compiler which libraries to use:
#ifdef _DEBUG
#pragma comment(lib, "opencv_imgproc320d.lib")
#pragma comment(lib, "opencv_core320d.lib")
#pragma comment(lib, "opencv_imgcodecs320d.lib")
#else
#pragma comment(lib, "opencv_imgproc320.lib")
#pragma comment(lib, "opencv_core320.lib")
#pragma comment(lib, "opencv_imgcodecs320.lib")
#endif // _DEBUG

using namespace Filter;
using namespace cv;
using namespace std;

Myriad::Myriad( Mat inInputImage , int inPatchSize , int inSearchSize , int inRefPatchCount , float inGamma )
:PatchSize( inPatchSize ) ,
        SearchSize( inSearchSize ) ,
        RefPatchCount( inRefPatchCount ) ,
        Gamma( inGamma ) ,
        InputImage( inInputImage )
{
    OutputImage = Mat::zeros( Size( InputImage.rows , InputImage.cols ) , CV_32F ); // output image should have the same dimension
    OutputGamma = Mat::zeros( Size( InputImage.rows , InputImage.cols ) , CV_32F ); // output image should have the same dimension
    int dimensions[] = { InputImage.rows , InputImage.cols , RefPatchCount + 1 };
    //PatchMatrix = Mat( 3 , dimensions , CV_32F );
    PatchMatrix = Mat::zeros( 3 , dimensions , CV_32F );
    InputImage.convertTo( InputImage , CV_32F );
}

Myriad::Myriad( Mat inPatchMatrix , float inGamma )
:PatchMatrix( inPatchMatrix ),
        Gamma( inGamma )
{
    OutputImage = Mat::zeros( PatchMatrix.size[ 0 ] , PatchMatrix.size[ 1 ] , CV_32F );
    OutputGamma = Mat::zeros( PatchMatrix.size[ 0 ] , PatchMatrix.size[ 1 ] , CV_32F );
    RefPatchCount = PatchMatrix.size[ 2 ] - 1;
    
    // the following values are rather arbitrarily initialized since they are not used when using THIS(!!!) constructor:
    SearchSize = 0;
    PatchSize = 0;
    InputImage = Mat::zeros( Size( 1 , 1 ) , CV_32F );
}

Myriad::Myriad( const Myriad& M )
:PatchSize( M.PatchSize ) ,
        SearchSize( M.SearchSize ) ,
        RefPatchCount( M.RefPatchCount ) ,
        Gamma( M.Gamma ) ,
        InputImage( M.InputImage ) ,
        OutputImage( M.OutputImage ) ,
        OutputGamma( M.OutputGamma ) ,
        PatchMatrix( M.PatchMatrix )
{
    // copy constructor
}

Myriad::~Myriad()
{
    
}

Myriad& Myriad::operator=( const Myriad& M )
{
    // assignment operator:
    
    this->PatchSize = M.PatchSize;
    this->SearchSize = M.SearchSize;
    this->RefPatchCount = M.RefPatchCount;
    this->Gamma = M.Gamma;
    this->InputImage = M.InputImage;
    this->OutputImage = M.OutputImage;
    this->OutputGamma = M.OutputGamma;
    this->PatchMatrix = M.PatchMatrix;
    
    return *this;
}

void Myriad::setGamma( float inGamma )
{
    Gamma = inGamma;
}

void Myriad::setPatchSize( int inPatchSize )
{
    PatchSize = inPatchSize;
}

void Myriad::setSearchSize( int inSearchSize )
{
    SearchSize = inSearchSize;
}

void Myriad::setRefPatchCount( int inRefPatchCount )
{
    RefPatchCount = inRefPatchCount;
}

//void Myriad::setInputFilename( std::string inInputFilename )
//{
//	FilenameInput = inInputFilename;
//}

//void Myriad::setOutputFilename( std::string inOutputFilename )
//{
//	FilenameOutput = inOutputFilename;
//}

void Myriad::setInputImage( Mat inInputImage )
{
    InputImage = inInputImage;
}

Mat Myriad::getOutputGamma()
{
    return OutputGamma;
}

Mat Myriad::getOutputImage()
{
    return OutputImage;
}

Mat Myriad::getPatchMatrix()
{
    return PatchMatrix;
}

void Myriad::PatchExtraction()
{
    //InputImage /= 255.0;
    
    Size size = InputImage.size();
    
    // main loop for generating patches, calls functions to compute its entries:
    int WindowHalf = cvFloor( float( PatchSize ) / 2.0 );
    int SearchHalf = cvFloor( float( SearchSize ) / 2.0 );
    
    float *DataPointer = (float *)(PatchMatrix.data);
    float *ValPtr;
    
    int height;
    int width;
    Mat Val;
    Mat ImagePatch;
    
    int numberOfPixels = size.width * size.height;
    
    for( int i = 0; i < InputImage.size[ 0 ]; ++i )
    {
        for( int j = 0; j < InputImage.size[ 1 ]; ++j )
        {
            //width = fmin( InputImage.cols - 1 , j + SearchHalf ) - fmax( 0 , j - SearchHalf ) + 1;
            //height = fmin( InputImage.rows - 1 , i + SearchHalf ) - fmax( 0 , i - SearchHalf ) + 1;
            
            //ImagePatch = InputImage( Rect( j - WindowHalf , i - WindowHalf , PatchSize , PatchSize ) );
            
            //Range rowrange( fmax( 0 , i - SearchHalf ) , fmin( InputImage.rows , i + SearchHalf + 1 ) );
            //Range colrange( fmax( 0 , j - SearchHalf ) , fmin( InputImage.cols , j + SearchHalf + 1 ) );
            //Mat SearchWindow = InputImage( rowrange , colrange );
            //Mat SearchWindow = InputImage( Rect( fmax( 0 , i - SearchHalf ) , fmax( 0 , j - SearchHalf ) , height , width ) );
            
            Val = FindPatches( i , j );
            
            ValPtr = Val.ptr<float>( 0 );
            
// TODO: if possible change data type of "PatchMatrix" to simple float vector< vector <float[]> > to save the following loop:
            
            int offset = j + size.width *i;
            for( int k = 0; k < RefPatchCount + 1; ++k )
            {
                DataPointer[ offset ] = ValPtr[ k ];
                offset += numberOfPixels;
            }
        }
    }
}


void Myriad::ComputeMyriadGeneralized( Mat Array, float * beta_out, float * gamma_out )
{
    // represents the main fixed point iteration:
    float beta = SelectionMyriad( Array );
    float gamma = 10.0f; // default initialization value.
    
    float *ArrayPtr;
    
    ArrayPtr = Array.ptr<float>( 0 ); // Pointer to row 0 of Array
    
    float eps = 1.0E-2;
    float dist = std::numeric_limits<float>::infinity();
    float beta_old;
    float gamma_old;
    float S0r;
    float S1r;
    int N = Array.cols;
    
    float toto;
    
     int i ;
    for( i = 0; i < 10000; ++i )
    {
        beta_old = beta;
        gamma_old = gamma;
        
        //beta = ComputeMyriadHelper( Array , beta );
        
        S0r = 0.0f;
        S1r = 0.0f;
        for( int j = 0; j < N; ++j ){
            
            float pre_calc = (ArrayPtr[ j ]-beta)/gamma;
            float pre_calc2 = pre_calc*pre_calc;
            S0r += 1/(1+pre_calc2);
            S1r += pre_calc/(1+pre_calc2);
        }
        S0r = S0r/N;
        S1r = S1r/N;
        
        beta = beta + gamma * S1r / S0r ;
        gamma = sqrt(gamma*gamma *(1/S0r - 1));
        
//         float signed_dist = (beta - beta_old)*(beta - beta_old)
//                          + (gamma - gamma_old)*(gamma - gamma_old);
        float signed_dist = beta - beta_old;
        float dist1 =  signed_dist > 0 ? signed_dist : -signed_dist;
        signed_dist = gamma - gamma_old;
        float dist2 =  signed_dist > 0 ? signed_dist : -signed_dist;
        if( dist1 + dist2 < eps )
        {
            break;
        }
    }
    
    //printf("iterations\t%d\n", i);
    
//     *beta_out=221.10f;
//     *gamma_out=5.52f;
    *beta_out=beta;
    *gamma_out=gamma;
    
}


void Myriad::ComputeGeneralized()
{
    // use InputImage size since patch matrix has same dimension for each layer.
    
    float *DataPointer = (float *)(PatchMatrix.data);
    
    float *ArrayPtr;
    
    float *OutputDataPointer;
    float *OutputGammaPointer;
    float val;
    
    Mat Array = Mat::zeros( 1 , PatchMatrix.size[ 2 ] , CV_32F );
    ArrayPtr = Array.ptr<float>( 0 ); // Pointer to row 0 of Array
    
    int numberOfPixels = PatchMatrix.size[ 1 ] * PatchMatrix.size[ 0 ];
    
    for( int i = 0; i < PatchMatrix.size[0]; ++i )
    {
        for( int j = 0; j < PatchMatrix.size[1]; ++j )
        {
            OutputDataPointer = OutputImage.ptr<float>( i );
            OutputGammaPointer = OutputGamma.ptr<float>( i );
            
            int offset = j + i * PatchMatrix.size[ 1 ];
            for( int k = 0; k < PatchMatrix.size[ 2 ]; ++k )
            {
                ArrayPtr[ k ] = DataPointer[ offset ];
                offset += numberOfPixels;
            }
            
            ComputeMyriadGeneralized( Array, &(OutputDataPointer[ j ]), &(OutputGammaPointer[ j ]) );
        }
    }
}

void Myriad::Compute()
{
    // use InputImage size since patch matrix has same dimension for each layer.
    
    float *DataPointer = (float *)(PatchMatrix.data);
    
    float *ArrayPtr;
    
    float *OutputDataPointer;
    float val;
    
    Mat Array = Mat::zeros( 1 , PatchMatrix.size[ 2 ] , CV_32F );
    ArrayPtr = Array.ptr<float>( 0 ); // Pointer to row 0 of Array
    
    int numberOfPixels = PatchMatrix.size[ 1 ] * PatchMatrix.size[ 0 ];
    
    for( int i = 0; i < PatchMatrix.size[0]; ++i )
    {
        for( int j = 0; j < PatchMatrix.size[1]; ++j )
        {
            OutputDataPointer = OutputImage.ptr<float>( i );
            
            int offset = j + i * PatchMatrix.size[ 1 ];
            for( int k = 0; k < PatchMatrix.size[ 2 ]; ++k )
            {
                ArrayPtr[ k ] = DataPointer[ offset ];
                offset += numberOfPixels;
            }
            
            OutputDataPointer[ j ] = ComputeMyriad( Array );
        }
    }
}

Mat Myriad::FindPatches( int xpos , int ypos )
{
    // compare each possible patch inside the search frame to find most similar patches:
    vector<float> SimList;
    vector<float> PatchMidPoint;
    
    int SearchHalf = cvFloor( float( SearchSize ) / 2.0 );
    
    float *InputImageDataPointer = (float *)(InputImage.data);
    
    // main loop for generating result image, calls functions to compute its entries:
    
    for( int i = fmax( 0 , xpos - SearchHalf ) ; i < fmin( InputImage.rows , xpos + SearchHalf ); ++i )
    {
        for( int j = fmax( 0 , ypos - SearchHalf ); j < fmin( InputImage.cols , ypos + SearchHalf ); ++j )
        {
            if( (i == xpos) && (j == ypos) )
                continue; // prevent the original patch from being added twice
            
            SimList.push_back( Similarity( xpos , ypos , i , j ) );
            
            PatchMidPoint.push_back( InputImageDataPointer[ j + i * InputImage.cols ] ); // extracts the reference point
        }
    }
    
    Mat Array = Mat::zeros( 1 , RefPatchCount + 1 , CV_32F );
    float* Arrayptr = Array.ptr<float>( 0 );
    Arrayptr[ 0 ] = InputImageDataPointer[ ypos + xpos * InputImage.cols ];
    for( int i = 1; i < RefPatchCount + 1; ++i )
    {
        Arrayptr[ i ] = VecMax( SimList , PatchMidPoint );
    }
    
    return Array;
}

float Myriad::VecMax( vector<float> &sim , vector<float> &vec)
{
    // iterate the input similarity vector to find maximal similarity, return corresponding patch element and delete both out of the vectors:
    float test = -std::numeric_limits<float>::infinity();
    int tmp;
    for( int i = 0; i < vec.size(); ++i )
    {
        if( sim[ i ] > test )
        {
            test = sim[ i ];
            tmp = i;
        }
    }
    
    float ret = vec[ tmp ];
    vec.erase( vec.begin() + tmp );
    sim.erase( sim.begin() + tmp );
    return ret;
}

float Myriad::ComputeMyriad( Mat Array )
{
    // represents the main fixed point iteration:
    float beta = SelectionMyriad( Array );
    float eps = 1.0E-2;
    float dist = std::numeric_limits<float>::infinity();
    float beta_old;
    
    //while( dist > eps )
    //{
    //	float beta_old = beta;
    //	beta = beta - ComputeMyriadHelper( Array , beta );
    //	dist = abs( beta - beta_old );
    //}
    int i;
    for( i = 0; i < 10000; ++i )
    {
        beta_old = beta;
        
        beta = ComputeMyriadHelper( Array , beta );
        
        float signed_dist = beta - beta_old;
        dist =  signed_dist > 0 ? signed_dist : -signed_dist;
        if( dist < eps )
        {
            break;
        }
    }
    
//    printf("iterations\t%d\n", i);
    
    return beta;
    //energy = SelectionMyriadHelper( Array , float x )
}


float Myriad::ComputeMyriadHelper( Mat Array , float x )
{
    //  computes the value of fixed point equation for the currently selected element x:
    
    //Array.copyTo( res );
    
    float *Arrayptr = Array.ptr<float>( 0 );
    
    float denom = 0, num = 0;
    float tmp;
    for( int j = 0; j < Array.cols; ++j )
    {
        tmp = 1.0f / (Gamma*Gamma + (Arrayptr[ j ] - x)*(Arrayptr[ j ] - x));
        num += Arrayptr[ j ] * tmp;
        denom += tmp;
    }
    float step = num / denom;
    
    
    /*float lambda = 0.2f;
     * float tmp=0.0f;
     * for( int j = 0; j < Array.cols; ++j )
     * {
     * tmp += (Arrayptr[ j ] - x) / (Gamma*Gamma + (Arrayptr[ j ] - x)*(Arrayptr[ j ] - x));
     * }
     * float step = x + lambda * tmp;*/
    
    
    return step;
}

float Myriad::SelectionMyriad( Mat Array )
{
    // represents the main loops and checks to find the patch value that minimizes the objective function:
    
    float beta =  std::numeric_limits<float>::infinity();
    float test;
    
    float *Arrayptr = Array.ptr<float>( 0 );
    float ret = Arrayptr[ 0 ];
    for( int j = 0; j < Array.cols; ++j )
    {
        test = SelectionMyriadHelper( Array , Arrayptr[ j ] );
        if( test < beta )
        {
            beta = test;
            ret = Arrayptr[ j ];
        }
    }
    
    //mexPrintf("%d\n", Array.cols);
    
    /* test mean*/
    /*ret=0.0f;
     * for( int j = 0; j < Array.cols; ++j )
     * {
     * ret += Arrayptr[ j ];
     * }
     * ret = ret/((float)(Array.cols));*/
    
    return ret; // TODO: add min(beta, median(Array))
}

float Myriad::SelectionMyriadHelper( Mat Array , float x )
{
    // compute each value of the objective function:
    float res = 0.0f;
    float tmp;
    float GammaSquare = Gamma*Gamma;
    float* Arrayptr = Array.ptr<float>( 0 );
    for( int i = 0; i < Array.cols; ++i )
    {
        //res *= GammaSquare / (tmp*tmp + GammaSquare);
        /*tmp = (Arrayptr[ i ] - x);
         * res *= ( tmp*tmp/GammaSquare + 1.0f);*/
        /*tmp = (Arrayptr[ i ] - x) / Gamma;
         * res *= 1.0f / (tmp*tmp + 1.0f);*/
        tmp = (Arrayptr[ i ] - x);
        res += log(tmp*tmp + GammaSquare);
        
    }
    return res;
}

//New Similarity measure function:
float Myriad::Similarity( int Initial_x , int Initial_y , int Reference_x , int Reference_y )
{
    float ans = 1.0f;
    float PatchHalf = float( PatchSize ) / 2.0f;
    
    // pointer to data:
    float *InitialDataPointer = (float *)(InputImage.data);
    
    /* compare patches*/
    for( int dy = -PatchHalf; dy <= PatchHalf; dy++ )
    {
        int ady = (PatchHalf + dy < InputImage.rows) ? Initial_y + dy : 2 * InputImage.rows - (Initial_y + dy) - 1;
        int bdy = (Reference_y + dy < InputImage.rows) ? Reference_y + dy : 2 * InputImage.rows - (Reference_y + dy) - 1;
        ady = (ady < 0) ? -ady : ady;
        bdy = (bdy < 0) ? -bdy : bdy;
        
        for( int dx = -PatchHalf; dx <= PatchHalf; dx++ )
        {
            int adx = (PatchHalf + dx < InputImage.cols) ? Initial_x + dx : 2 * InputImage.cols - (Initial_x + dx) - 1;
            int bdx = (Reference_x + dx < InputImage.cols) ? Reference_x + dx : 2 * InputImage.cols - (Reference_x + dx) - 1;
            adx = (adx < 0) ? -adx : adx;
            bdx = (bdx < 0) ? -bdx : bdx;
            
            float ac = InitialDataPointer[ ady + InputImage.cols * adx ];
            float bc = InitialDataPointer[ bdy + InputImage.cols * bdx ];
            
            ans *= 1.0f / ((ac - bc)*(ac - bc) / (Gamma*Gamma*2.0f*2.0f) + 1.0f);
        }
    }
    
    return ans;
}

float Myriad::Similarity( Mat ImagePatch , Mat ReferencePatch )
{
    // computes the similarity of two input patches:
    
    float res = 1.0f;
    
    float *Imageptr;
    float *Refptr;
    float tmp;
    
    for( int j = 0; j < ImagePatch.rows; ++j )
    {
        Imageptr = ImagePatch.ptr<float>( j );
        Refptr = ReferencePatch.ptr<float>( j );
        for( int i = 0; i < ImagePatch.cols; ++i )
        {
            /*tmp = (Imageptr[ i ] - Refptr[ i ]) / (Gamma * 2.0f);
             * res *= 1.0f / (tmp*tmp + 1.0f);*/
            tmp = (Imageptr[ i ] - Refptr[ i ]) / (2.0f);
            float gamma_square = Gamma*Gamma;
            res *= gamma_square / (tmp*tmp + gamma_square);
        }
    }
    
    return res;
}
