/* Distance de similarité sous hypothèse de bruit Cauchy */

#include "matrix.h"
#include "mex.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits>


/*
 * similarity.cpp
 *
 * Auteur : Fabien Pierre
 * Date : 20/06/2017
 *
 *
 * Compilation Linux with Matlab linking:
 * mex  -largeArrayDims cauchy_similarity.cpp -output cauchy_similarity
 *
 * input : image gray scale single precision NxM
 * gamma (parameter of Cauchy estimator)
 * K (number of retained patches)
 * patch radius
 * search window radius
 *
 * output : single precision matrix, size NxMxK
 * for i,j, output(i,j,:) is the vector of the K nearest patch center
 *
 */


class FLOATMAP { public:
    int w, h;
    float *data;
    FLOATMAP(int w_, int h_) :w(w_), h(h_) { data = new float[w*h]; }
    FLOATMAP(int w_, int h_, float *data_) :w(w_), h(h_), data(data_) {}
    ~FLOATMAP() {  }
    float *operator[](int y) { return &data[y*w]; }
};


/*Global variables*/
int patch_w;
int W_radius;
int patch_radius;
float gamma_param;
int K;

int nb_log_cpt;

/*Define the Cauchy similarity. */
float cauchy_similarity(float & a, float & b){
    float gammacarre = gamma_param*gamma_param*4.0f;
    float ret = -log((a-b)*(a-b) + gammacarre);
    return ret;
}


/* Measure distance between 2 patches
 * with center point (ax, ay) and (bx, by)
 ****Based on Cauchy similarity
 */

float * distances_temp;
float * distances;
int * nbValidSamples;

int compare (const void * b, const void * a)
{
    float to_test =  distances_temp[*(int*)a] - distances_temp[*(int*)b] ;
    
    if(to_test>0)
        return 1;
    else if(to_test<0)
        return -1;
    else
        return 0;
}

int iabs(int a){
    return (a > 0? a :-a );
}

/* search of the K neearest neighbors */
void KNN(FLOATMAP *a,  float * k_best_centers, float * poids) {
    
    int width  = a->w ;
    int height = a->h ;
    
    /*initialize variables*/
    int nb_pix = width*height;
    int W_width = 2*W_radius+1;
    int pixel_arround_in_window = W_width*W_width;
    distances = (float*) malloc(nb_pix*pixel_arround_in_window*sizeof(float));
    nbValidSamples = (int*) malloc(nb_pix*sizeof(int));
    
    int * indexes = (int*) malloc(nb_pix*pixel_arround_in_window*sizeof(int));
    int * idbx    = (int*) malloc(nb_pix*pixel_arround_in_window*sizeof(int));
    int * idby    = (int*) malloc(nb_pix*pixel_arround_in_window*sizeof(int));
    
    
    for(int i=0; i < pixel_arround_in_window; i++){
        distances[i]=0.0f;
        indexes[i]=i;
    }
    
    int aew = a->w;
    int aeh = a->h;
    
    float MINFINITY;
    MINFINITY= -std::numeric_limits<float>::infinity();
    for (int ay = 0; ay < aeh; ay ++) {
        for (int ax = 0; ax < aew; ax ++) {
            
            for(int i=0; i< pixel_arround_in_window; i++){
                distances[i + pixel_arround_in_window*(ax+ ay*width)] = MINFINITY;
                indexes[i + pixel_arround_in_window*(ax+ ay*width)] = i;
            }
            
            
        }
    }
    
    int number_pixel_window ;
    
    
    /* compare patches*/
    for (int ay = 0; ay < aeh; ay ++) {/*browse image pixels*/
        for (int ax = 0; ax < aew; ax ++) {/*browse image pixels*/
            
            /*define limit of the search window*/
            int ystart_windows = ay-W_radius < 0 ? 0 : ay-W_radius;
            int yend_windows = ay+W_radius >= aeh ? aeh-1 : ay+W_radius;
            int xstart_windows = ax-W_radius < 0 ? 0 : ax-W_radius;
            int xend_windows = ax+W_radius >= aew ? aew-1 : ax+W_radius;
            
            
            /*computation of distances*/
            for  (int by = ystart_windows; by <= yend_windows; by ++) {/*browse search window*/
                for (int bx = xstart_windows; bx <= xend_windows; bx ++) {/*browse search window*/
                    
                    
                    float ac = (*a)[ay][ax];
                    float bc = (*a)[by][bx];
                    float res = cauchy_similarity(ac, bc);
                    
                    
                    for (int dy = -patch_radius; dy <= patch_radius; dy++) {/*browse patch*/
                        
                        int ady_init = ay+dy;
                        int ady= (ady_init < a->h)? ady_init: 2*a->h-(ady_init)-1;
                        ady= (ady < 0)? -ady-1 : ady;
                        int bdy_init = by+dy;
                        int bdy= (bdy_init < a->h)? bdy_init: 2*a->h-(bdy_init)-1;
                        bdy= (bdy < 0)? -bdy-1 : bdy;
                        
                       int offsetmin = (iabs(ady-ady_init) > iabs(bdy-bdy_init)) ? ady-ady_init : bdy-bdy_init;
                        ady = ady_init + offsetmin;
                        bdy = bdy_init + offsetmin;
                        
                        
                        
                        
                        int off_y = (bdy-ady+W_radius) * W_width;
                        
                        for (int dx = -patch_radius; dx <= patch_radius; dx++) {/*browse patch*/
                            
                            int adx_init = ax+dx;
                            int adx = (adx_init < a->w) ? adx_init :  2*a->w-(adx_init)-1;
                            adx = (adx < 0)? -adx-1 : adx;
                            int bdx_init = bx+dx;
                            int bdx = (bdx_init < a->w) ? bdx_init :  2*a->w-(bdx_init)-1;
                            bdx = (bdx < 0)? -bdx-1 : bdx;
                            
                        
                       int offsetminx = (iabs(adx-adx_init) > iabs(bdx-bdx_init)) ? adx-adx_init : bdx-bdx_init;
                        adx = adx_init + offsetminx;
                        bdx = bdx_init + offsetminx;
                            
                            
                            number_pixel_window = (bdx-adx+W_radius)  + off_y;
                            
// mexPrintf("ay=%d\tax=%d\tby=%d\tbx=%d\tady=%d\tadx=%d\tbdy=%d\tbdx=%d\n", ay, ax, by, bx, ady, adx, bdy, bdx); 
                            
                            int index_precomputed = number_pixel_window + pixel_arround_in_window*(adx+ ady*width);
                            
                            if (distances[index_precomputed] == MINFINITY){
                                distances[index_precomputed] = res;
                            }
                            else{
                                distances[index_precomputed] += res;
                            }
                            
                        }
                    }
                    
                    number_pixel_window = (bx-ax+W_radius)  + (by-ay+W_radius) * W_width;
                    int offset_a = pixel_arround_in_window*(ax+ ay*width);
                    
                    idbx[number_pixel_window + offset_a]= bx;
                    idby[number_pixel_window + offset_a]= by;
                }
            }
            
        }
    }
    
    
                        
            
    
    for (int ay = 0; ay < aeh; ay ++) {/*browse image pixels*/
        for (int ax = 0; ax < aew; ax ++) {/*browse image pixels*/
            /*sort indexes*/
            distances_temp   = &(distances[0+pixel_arround_in_window*(ax+ ay*width)]);
            int *indexes_tmp = &(  indexes[0+pixel_arround_in_window*(ax+ ay*width)]);
            qsort (indexes_tmp, pixel_arround_in_window, sizeof(int), compare);
            
            /*copy the K best patch centers.*/
            
            for(int id_k=0; id_k < K; id_k++){
                k_best_centers[ax+ ay*width + id_k*height*width] =  (*a)
                [idby[indexes_tmp[id_k+1]+pixel_arround_in_window*(ax+ ay*width)]]
                        [idbx[indexes_tmp[id_k+1]+pixel_arround_in_window*(ax+ ay*width)]];
						
				poids[ax+ ay*width + id_k*height*width] = distances_temp[indexes_tmp[id_k+1]];
            }
        }
    }
    
    
    
    
//     for (int ay = 0; ay < aeh; ay ++) {/*browse image pixels*/
//         for (int ax = 0; ax < aew; ax ++) {/*browse image pixels*/
//             int *indexes_tmp = &(  indexes[0+pixel_arround_in_window*(ax+ ay*width)]);
//             // mexPrintf("%d\t%d\t\t", ay, ax);
//             for(int id_k=0; id_k < pixel_arround_in_window; id_k++){
//                 // mexPrintf("%lf\t", distances[indexes_tmp[id_k]+pixel_arround_in_window*(ax+ ay*width)]);
//             }
//             // mexPrintf("\n");
//         }
//     }
    
    for (int ay = 0; ay < aeh; ay ++) {/*browse image pixels*/
        for (int ax = 0; ax < aew; ax ++) {/*browse image pixels*/
            int *indexes_tmp = &(  indexes[0+pixel_arround_in_window*(ax+ ay*width)]);
            // mexPrintf("%d\t%d\t\t", ay, ax);
            for(int id_k=0; id_k < pixel_arround_in_window; id_k++){
                // mexPrintf("%lf\t", distances[id_k+pixel_arround_in_window*(ax+ ay*width)]);
            }
            // mexPrintf("\n");
        }
    }
    
            
    
    
    free(distances);
    free(indexes);
    free(idbx);
    free(idby);
    
    return;
}



void mexFunction (int nlhs, mxArray *plhs[],
        int nrhs, const mxArray *prhs[])
{
    /* Vérification du nombre d'arguments d'entrée :
     * image input , gamma, K, patch width, search window radius
     **/
    if (nrhs != 5)
    {
        mexErrMsgIdAndTxt("KNNS:nrhs","5 inputs required");
    }
    
	
    if (nlhs != 2)
    {
        mexErrMsgIdAndTxt("KNNS:nlhs","2 output required");
    }
    
    // mexPrintf("\n--------------------------\n");
    // mexPrintf("Call KNN similarity Cauchy\n");
    // mexPrintf("--------------------------\n");
    
// Gestion entree-sortie
    
    const size_t * dims = (const size_t*)(mxGetDimensions(prhs[0]));
    int dim = mxGetNumberOfDimensions(prhs[0]);
    
    int wA= (int)dims[0];
    int hA = (int)dims[1];
    
    
    patch_w = (int)(mxGetScalar(prhs[3]));
    patch_radius = (patch_w-1)/2;
    gamma_param= (float)(mxGetScalar(prhs[1]));
    
    W_radius = (int)(mxGetScalar(prhs[4]));
    K = (int)(mxGetScalar(prhs[2]));
    
    /* penser à vérifier en entrée que */
    if (((W_radius+1)*(W_radius+1))-1 < K)
    {
        mexWarnMsgIdAndTxt("KNNS:checkParam",
                "K is too large in comparison of window radius.\nFrom now K=%d\n",
                ((W_radius+1)*(W_radius+1))-1);
        K=((W_radius+1)*(W_radius+1))-1;
    }
    
    const int n_dim_result= 3;
    size_t* dims_result = (size_t *)(malloc(sizeof(size_t)*n_dim_result));
    dims_result[0]= (size_t)wA;
    dims_result[1]= (size_t)hA;
    dims_result[2]= (size_t)K;
    
    plhs[0]= mxCreateNumericArray(n_dim_result, dims_result ,
            mxSINGLE_CLASS, mxREAL     ) ;
    plhs[1]= mxCreateNumericArray(n_dim_result, dims_result ,
            mxSINGLE_CLASS, mxREAL     ) ;
    
    FLOATMAP *A = new FLOATMAP (wA,hA,(float*)(mxGetPr(prhs[0])));
    
    // mexPrintf("size image:\t %d x %d \ngamma=\t\t %lf \n", wA,hA,gamma_param);
    // mexPrintf("K= \t\t\t%d \npatch width=\t\t%d\n",  K,  2*patch_radius+1);
    // mexPrintf("search window width=\t%d\n", 2*W_radius+1  );
    // mexPrintf("--------------------------\n\n");
    
    KNN(A,  (float*)(mxGetPr(plhs[0])),  (float*)(mxGetPr(plhs[1])));
    
    delete A;
    
    return;
}
